import { TokenType } from "./tokentype";
import { ASTNode } from "./definition";
declare class Parser {
    input: string;
    pos: number;
    start: number;
    end: number;
    type: TokenType;
    value: string | undefined;
    lastTokStart: number;
    lastTokEnd: number;
    constructor(query: string);
    /**
     * Exception handling
     */
    raise(pos: number, msg: string): never;
    next(): void;
    nextToken(): void;
    readToken(code: number): void;
    fullCharCodeAtPos(): number;
    skipSpace(): void;
    readTokenSlash(): void;
    getTokenFromCode(code: number): void;
    finishToken(type: TokenType, val?: string): void;
    /**
     * Read an integer in the given radix. Return null if zero digits
     * were read, the integer value otherwise. When `len` is given, this
     * will return `null` unless the integer has exactly `len` digits.
     */
    readInt(radix: number, len?: null): number | null;
    readNumber(): void;
    readString(quote: number): void;
    readEscapedChar(): string;
    readWord1(): string;
    readWord(): void;
    startNode(): ASTNode;
    finishNode(node: ASTNode, type: string): ASTNode;
    parse(): ASTNode[];
    parseTopLevel(): ASTNode[];
    parseStep(): ASTNode;
    parseAxis(): TokenType;
    parseNodeTest(): ASTNode;
    parseTag(node: ASTNode): ASTNode;
    parseFilters(): ASTNode[];
    parsePredicate(): ASTNode;
    parseAttrFilter(node: ASTNode): ASTNode;
    parseFunction(node: ASTNode): ASTNode;
    eat(type: TokenType): boolean;
    expect(type: TokenType): boolean;
    getName(): string | undefined;
    getString(): string | undefined;
    parseBindingList(close: TokenType): ASTNode[];
    parseArg(): ASTNode;
    parseMaybeFun(node: ASTNode): ASTNode;
}
export default Parser;
