"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const tokentype_1 = __importDefault(require("./tokentype"));
function axisToStr(axis, isFirst) {
    if (isFirst) {
        return "";
    }
    if (axis === tokentype_1.default.doubleSlash) {
        // child or descendent
        return " ";
    }
    else if (axis === tokentype_1.default.slash) {
        return " > ";
    }
    throw Error("!Unexpected axis");
}
function nodeTestToStr(node) {
    if (node.value === "*") {
        return "";
    }
    return node.value || "";
}
function attrFilterToStr(node) {
    const attr = node.attr;
    if (attr === "id") {
        return `#${node.value}`;
    }
    else if (attr === "class") {
        return `.${node.value}`;
    }
    else {
        return `[${attr}="${node.value}"]`;
    }
}
function xpathFuncToStr(node) {
    switch (node.name) {
        case "contains":
            return containsToStr(node);
        case "position":
            return positionToStr(node);
        default:
            throw new Error(`Unsupported function: ${node.name}`);
    }
}
function positionToStr(node) {
    return `:nth-child(${node.position})`;
}
function containsToStr(node) {
    if (!node.arg) {
        throw new Error(`No arguments for function: ${node.name}`);
    }
    if (node.arg.length !== 2) {
        throw new Error("Expect contains function to have 2 arguments");
    }
    if (node.arg[0].name === "text") {
        return `:contains(${node.arg[1].name})`;
    }
    if (node.arg[0].type === "attribute") {
        return `[${node.arg[0].attr}*=${node.arg[1].name}]`;
    }
    throw new Error(`Unsupported function: ${node.arg[0].type}`);
}
function predicateToStr(node) {
    switch (node.type) {
        case "positionFilter":
            return `:nth-child(${node.position})`;
        case "attributeFilter":
            return attrFilterToStr(node);
        case "function":
            return xpathFuncToStr(node);
        case "name":
            return `:has(${node.name})`;
        default:
            throw new Error("!Unhandled predicate");
    }
}
function stepToSel(step, index) {
    const axis = axisToStr(step.axis, index === 0);
    const nodeTest = nodeTestToStr(step.nodeTest);
    const predicate = (step.predicate || []).map(predicateToStr);
    return axis + nodeTest + predicate.join("");
}
function xpath2css(steps) {
    const selectorList = steps.map(stepToSel);
    return selectorList.join("");
}
exports.default = xpath2css;
